#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <getopt.h>
#include <omp.h>
#include "fire.h"

int forest_size;
int trees;

void usage() {
    fprintf(stderr,"Usage: %s -p <burn-probability> -n <size-forest>, where\n \
\t[-p float] probability that a green tree will catch fire due to burning neighnor, between 0 and 1\n \
\t[-n int] the number of trees per side of a square forest, between 2 and 80\n",
            program_invocation_short_name);
}

struct user_args get_user_args(const int argc,char **argv) {
    int opt, tmp;
    struct user_args args;

    args.burn_ratio = 0;
    while((opt = getopt(argc,argv,":p:n:")) != -1) {
        switch(opt) {
            case 'p':
                args.burn_ratio = atof(optarg);
                if(args.burn_ratio == HUGE_VAL) {
                    fprintf(stderr,"Received invalid value: %s, %s:%d\n",
                            strerror(errno),__FILE__,__LINE__);
                    usage();
                    abort();
                }
               break; 
	    case 'n':
		tmp = atof(optarg);
		if(tmp>80 || tmp<2){
                    fprintf(stderr,"Received invalid value: %s, %s:%d\n",
                            strerror(errno),__FILE__,__LINE__);
                    usage();
                    abort();
		}
		forest_size = tmp;
		trees = forest_size*forest_size;
		break;
	    default:
                fprintf(stderr,"Received invalid flag: %s, %s:%d\n",strerror(errno),__FILE__,__LINE__);
		usage();
		abort();
        }
    }

    if(args.burn_ratio <= 0 || args.burn_ratio > 1) {
        fprintf(stderr,"Supply valid burn ratio!\n");
        usage();
        abort();
    }

    return args;
}

struct forest *malloc_forest(
        const struct user_args args,
        struct drand48_data *rand_buffer) {

    struct forest *burning_forest;
    size_t forest_bytes = sizeof(forest) * trees;
    double rand_result;
    unsigned int i;
    int up,down,right,left;
    
    if((burning_forest = malloc(forest_bytes)) == NULL) {
        fprintf(stderr,"Failed to malloc %zd bytes for forest: %s, %s:%d\n", forest_bytes,strerror(errno),__FILE__,__LINE__);
	return NULL;
    }

    for(i=0;i<trees;i++){
	burning_forest[i].tree = GREEN_TREE;

	up = i-forest_size;
	down = i-forest_size;
        left = i-1;
        right = i+1;

        burning_forest[i].neighbors.up = up >= 0
            ? up
            : OOB;

        burning_forest[i].neighbors.left = (left >= 0 && i%forest_size != 0)
            ? left
            : OOB;

        burning_forest[i].neighbors.down = down < trees
            ? down
            : OOB;

        burning_forest[i].neighbors.right = (right < trees && i%forest_size != forest_size-1)
            ? right
            : OOB;
    }

    drand48_r(rand_buffer,&rand_result);
    i = (int)(rand_result*trees);

    burning_forest[i].tree++;
    
    return burning_forest;
}

struct forest *init_forest(
        const struct user_args args,
        struct drand48_data *rand_buffer) {
    struct forest *burning_forest;
    size_t forest_bytes = sizeof(forest)*trees;
    double rand_result;
    unsigned int i;
    int up,down,right,left;

    if((burning_forest = malloc(forest_bytes)) == NULL) {
        fprintf(stderr,"Failed to malloc %zd bytes for forest: %s, %s:%d\n", forest_bytes,strerror(errno),__FILE__,__LINE__);
        return NULL;
    }

    /* Set each cell to be a fresh tree */
    for(i=0;i<trees;i++) {
        burning_forest[i].tree = GREEN_TREE;

        /* Cache neighbor information */
        up = i-forest_size;
        down = i+forest_size;
        left = i-1;
        right = i+1;

        burning_forest[i].neighbors.up = up >= 0
            ? up
            : OOB;

        burning_forest[i].neighbors.left = (left >= 0 && i%forest_size != 0)
            ? left
            : OOB;

        burning_forest[i].neighbors.down = down < trees
            ? down
            : OOB;

        burning_forest[i].neighbors.right = (right<trees && i%forest_size!=forest_size-1)
            ? right
            : OOB;
    }

    drand48_r(rand_buffer,&rand_result);
    i = (int)(rand_result*trees);

    /* Increment by one to start it burning */
    burning_forest[i].tree++;

    return burning_forest;
}

int burn_trees(
        const struct user_args args,
        struct forest *burning_forest,
        struct forest *next_burning_forest,
        struct drand48_data *rand_buffer
        ) {
   unsigned int i,j,t,cur_burn_trees,started_burn_trees; 

   /* The simulation is over when no burning trees remain */
   t = 0;
   do {
	cur_burn_trees = started_burn_trees = 0;
	for(i=0;i<trees;i++) {
		burning_forest[i].tree = next_burning_forest[i].tree;
	}
	for(i=0;i<trees;i++) {
		if(burning_forest[i].tree > GREEN_TREE && burning_forest[i].tree < BLACK_TREE) {
                	next_burning_forest[i].tree++;
                	cur_burn_trees++;
                	started_burn_trees += burn_neighbors(args,burning_forest,next_burning_forest,i,rand_buffer);
		}
	}

    // Each thread runs the block below, one at a time.
#pragma omp critical
    {
        if(t==0){
            fprintf(stderr,"Thread:%d Time:0, 1 tree is on fire\n",
                    omp_get_thread_num());
        } else {
            fprintf(stderr, "Thread:%d Time:%u, %u trees started burning this \
time step, %u were already burning\n", omp_get_thread_num(), t,
                    started_burn_trees,cur_burn_trees);
        }

        t++;

        for(i=0;i<forest_size+1;i++){fprintf(stderr, "__");}
        fprintf(stderr,"_\n");

        for(i=0;i<forest_size;i++){
            fprintf(stderr, "| ");
            for(j=0;j<forest_size;j++){
                fprintf(stderr,"%u ",burning_forest[i*forest_size+j].tree);
            }
            fprintf(stderr,"|\n");
        }

        for(i=0;i<2*forest_size+3;i++){fprintf(stderr, "-");}
        fprintf(stderr,"\n");
    }

    usleep(100000);/* This slows the display of the forest to make it look like animation */


   } while((cur_burn_trees+started_burn_trees) > 0);

   return 0;
}

unsigned short int burn_neighbors(
        const struct user_args args,
        struct forest *burning_forest,
        struct forest *next_burning_forest,
        const unsigned int i,
        struct drand48_data *rand_buffer
        ) {
    unsigned short int started_burn_trees = 0;
    double rand_result;

    if(burning_forest[i].neighbors.up != OOB && burning_forest[burning_forest[i].neighbors.up].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
            if(next_burning_forest[burning_forest[i].neighbors.up].tree==0){
                started_burn_trees++;
            }
            next_burning_forest[burning_forest[i].neighbors.up].tree=1;
        }
    }

    if(burning_forest[i].neighbors.down != OOB
            && burning_forest[burning_forest[i].neighbors.down].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
            if(next_burning_forest[burning_forest[i].neighbors.down].tree==0){
                started_burn_trees++;
            }
            next_burning_forest[burning_forest[i].neighbors.down].tree=1;
        }
    }

    if(burning_forest[i].neighbors.left != OOB
            && burning_forest[burning_forest[i].neighbors.left].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
            if(next_burning_forest[burning_forest[i].neighbors.left].tree==0){
                started_burn_trees++;
            }
            next_burning_forest[burning_forest[i].neighbors.left].tree=1;
        }
    }
    if(burning_forest[i].neighbors.right != OOB
            && burning_forest[
            burning_forest[i].neighbors.right].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
            if(next_burning_forest[burning_forest[i].neighbors.right].tree==0){
                started_burn_trees++;
            }
            next_burning_forest[burning_forest[i].neighbors.right].tree=1;
        }
    }

    return started_burn_trees;
}

